/*
 * Copyright 2008-2010 Fabien Coeurjoly <fabien.coeurjoly@wanadoo.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  The names of contributors may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 */

#include <proto/asl.h>
#include <clib/macros.h>
#include <proto/dos.h>
#include <workbench/workbench.h>

#include "metadata.h"
#include "morphos_stuff.h"
#include "gui/interface.h"
#include "gui.h"


#ifdef __amigaos4__
static LONG AppMsgFunc (struct Hook *h, Object *obj, struct AppMessage **x)
{
#else
static LONG AppMsgFunc(void)
{
	struct AppMessage **x = (struct AppMessage **) REG_A1;
        APTR obj = (APTR) REG_A2;
#endif

	struct WBArg *ap;
	struct AppMessage *amsg = *x;
	int i;
	char buf[256];

	for (ap=amsg->am_ArgList,i=0;i<amsg->am_NumArgs;i++,ap++)
	{
		if(NameFromLock(ap->wa_Lock, buf, sizeof(buf)))
		{
			AddPart(buf, ap->wa_Name, sizeof(buf));

			if(!parse_filename(buf, playtree, mconfig, 0))
			{
				BPTR l;

				l = Lock(buf, ACCESS_READ);

				if(l)
				{
					struct FileInfoBlock fi;

					if(Examine(l, &fi))
					{
						if(fi.fib_DirEntryType > 0)
						{
							adddirtoplaylist(mygui->playlist, buf, TRUE, TRUE);
						}
						else
						{
							mygui->playlist->add_track(mygui->playlist, buf, NULL, NULL, 0, STREAMTYPE_FILE);
							DoMethod(obj, MM_PlaylistGroup_Add, mygui->playlist->tracks[mygui->playlist->trackcount - 1]);
						}
					}
					UnLock(l);
				}
			}
		}
	}

	return(0);
}

#ifdef __amigaos4__
static struct Hook AppMsgHook = { {0, 0}, (HOOKFUNC)&AppMsgFunc, NULL, NULL };
#else
static struct EmulLibEntry AppMsgHookGate = { TRAP_LIB, 0, (void (*)(void))AppMsgFunc };
static struct Hook AppMsgHook = { {0, 0}, (HOOKFUNC)&AppMsgHookGate, NULL, NULL };
#endif


/******************************************************************
 * playlistgroupclass
 *****************************************************************/

struct PlaylistGroupData
{
	APTR LI_Playlist;
	APTR BT_Play;
	APTR BT_Add;
	APTR BT_Remove;
	APTR BT_MoveUp;
	APTR BT_MoveDown;
	APTR BT_Clear;
	APTR BT_Shuffle;

	APTR BT_Save;

	char last_added[MAX_PATH];
};

DEFNEW(PlaylistGroup)
{
	APTR LI_Playlist, BT_Play, BT_Add, BT_Remove, BT_MoveUp, BT_MoveDown, BT_Save, BT_Shuffle, BT_Clear;

	obj = DoSuperNew(cl, obj,
		MUIA_Group_Horiz, TRUE,
		Child,
			LI_Playlist = NewObject(getplaylistlistclass(), NULL, TAG_DONE),

		Child,
			VGroup, MUIA_Weight, 0,
				Child, BT_Play     = MakeButton("Play"),
				Child, BT_Add      = MakeButton("Add..."),
				Child, BT_Remove   = MakeButton("Remove"),
				Child, BT_Clear    = MakeButton("Remove All"),
				Child, BT_MoveUp   = MakeButton("Move Up"),
				Child, BT_MoveDown = MakeButton("Move Down"),
				Child, BT_Shuffle  = MakeButton("Shuffle"),
				Child, VSpace(0),
				Child, BT_Save     = MakeButton("Save As..."),
				End,

		TAG_DONE
	);

	if (obj)
	{
		struct PlaylistGroupData *data = INST_DATA(cl, obj);

		data->LI_Playlist = LI_Playlist;
		data->BT_Play = BT_Play;
		data->BT_Add = BT_Add;
		data->BT_Remove = BT_Remove;
		data->BT_Clear = BT_Clear;
		data->BT_MoveUp = BT_MoveUp;
		data->BT_MoveDown = BT_MoveDown;
		data->BT_Shuffle = BT_Shuffle;
		data->BT_Save = BT_Save;

		/* buttons notifications */

		DoMethod(data->BT_Play, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_Play);

		DoMethod(data->BT_Add, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_AddRequester);

		DoMethod(data->BT_Remove, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_Remove);

		DoMethod(data->BT_Clear, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_RemoveAll);

		DoMethod(data->BT_MoveUp, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_MoveUp);

		DoMethod(data->BT_MoveDown, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_MoveDown);

		DoMethod(data->BT_Shuffle, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_Shuffle);

		DoMethod(data->BT_Save, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_PlaylistGroup_SaveRequester);

		DoMethod(data->LI_Playlist, MUIM_Notify, MUIA_Listview_DoubleClick, MUIV_EveryTime,
				 obj, 1, MM_PlaylistGroup_Play);

		/* Handle appwindow events */
		DoMethod(obj, MUIM_Notify, MUIA_AppMessage, MUIV_EveryTime,
				 obj, 3, MUIM_CallHook, &AppMsgHook, MUIV_TriggerValue);

		/* Restore if needed */

		if(remember_playlist)
		{
			DoMethod(obj, MM_PlaylistGroup_Load, "PROGDIR:conf/lastplaylist.pls");
		}
	}

	return (ULONG)obj;
}

DEFDISP(PlaylistGroup)
{
	if(remember_playlist)
	{
		DoMethod(obj, MM_PlaylistGroup_Save, "PROGDIR:conf/lastplaylist.pls");
	}
	return (DOSUPER);
}

DEFGET(PlaylistGroup)
{
	switch (msg->opg_AttrID)
	{
		case MA_PlaylistGroup_Current:
		{
			*msg->opg_Storage = mygui->playlist->current;
		}
		return (TRUE);
	}

	return DOSUPER;
}

static void doset(APTR obj, struct PlaylistGroupData *data, struct TagItem *tags)
{
	FORTAG(tags)
	{
		case MA_PlaylistGroup_Current:
			mygui->playlist->current = tag->ti_Data;
			break;
	}
	NEXTTAG
}

DEFSET(PlaylistGroup)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	doset(obj, data, INITTAGS);

	return DOSUPER;
}

DEFSMETHOD(PlaylistGroup_Add)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	DoMethod(data->LI_Playlist, MUIM_List_InsertSingle, msg->entry, MUIV_List_Insert_Bottom);

	return 0;
}

DEFTMETHOD(PlaylistGroup_AddRequester)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	APTR tags[] = { (APTR) ASLFR_TitleText, (APTR) "Select Files...",
					(APTR) ASLFR_InitialPattern, (APTR) MorphOS_GetPattern(), //"("PATTERN_STREAM "|" PATTERN_PLAYLIST")",
				    TAG_DONE };

	char ** files = NULL;
	STRPTR ptr;
	ULONG count;

	ptr = data->last_added[0] ? data->last_added : MorphOS_GetLastVisitedPath();


	set(app, MUIA_Application_Sleep, TRUE);

	count = asl_run_multiple(ptr, (APTR)&tags, &files, TRUE);

	set(app, MUIA_Application_Sleep, FALSE);

	if (files)
	{
		ULONG i;

		// Enqueue others
		for(i = 0; i < count; i++)
		{
			if(!parse_filename(files[i], playtree, mconfig, 0))
			{
				BPTR l;

				l = Lock(files[i], ACCESS_READ);

				if(l)
				{
					struct FileInfoBlock fi;

					if(Examine(l, &fi))
					{
						if(fi.fib_DirEntryType > 0)
						{
							adddirtoplaylist(mygui->playlist, files[i], TRUE, TRUE);
						}
						else
						{
							mygui->playlist->add_track(mygui->playlist, files[i], NULL, NULL, 0, STREAMTYPE_FILE);
							DoMethod(obj, MM_PlaylistGroup_Add, mygui->playlist->tracks[mygui->playlist->trackcount - 1]);
						}

						stccpy(data->last_added, files[i], sizeof(data->last_added));
					}
					UnLock(l);
				}
			}
		}

		update_playlistwindow(TRUE);

		asl_free(count, files);
	}

	return 0;
}

DEFTMETHOD(PlaylistGroup_Remove)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	/* don't remove currently played entry */
	if(mygui->playlist->current == getv(data->LI_Playlist, MUIA_List_Active))
	{
		return 0;
	}

	mygui->playlist->remove_track(mygui->playlist, getv(data->LI_Playlist, MUIA_List_Active) + 1);

	DoMethod(data->LI_Playlist, MUIM_List_Remove, MUIV_List_Remove_Active);

	return 0;
}

DEFTMETHOD(PlaylistGroup_RemoveAll)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	char * filename = mygui->playlist->tracks[(mygui->playlist->current)]->filename ? strdup(mygui->playlist->tracks[(mygui->playlist->current)]->filename) : NULL;
	char * title = mygui->playlist->tracks[(mygui->playlist->current)]->title ? strdup(mygui->playlist->tracks[(mygui->playlist->current)]->title) : NULL;
	char * artist = mygui->playlist->tracks[(mygui->playlist->current)]->artist ? strdup(mygui->playlist->tracks[(mygui->playlist->current)]->artist) : NULL;
	int duration = mygui->playlist->tracks[(mygui->playlist->current)]->duration;
	int type = mygui->playlist->tracks[(mygui->playlist->current)]->type;

	DoMethod(data->LI_Playlist, MUIM_List_Clear);

	mygui->playlist->clear_playlist(mygui->playlist);
	mygui->playlist->add_track(mygui->playlist, filename, title, artist, duration, type);
	mygui->playlist->current = 0;

	DoMethod(obj, MM_PlaylistGroup_Refresh, TRUE);

	free(filename);
	free(title);
	free(artist);

	return 0;
}

DEFTMETHOD(PlaylistGroup_MoveUp)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	mygui->playlist->moveup_track(mygui->playlist, getv(data->LI_Playlist, MUIA_List_Active) + 1);

	DoMethod(data->LI_Playlist, MUIM_List_Move, MUIV_List_Move_Active, MUIV_List_Move_Previous);

	set(data->LI_Playlist, MUIA_List_Active, MUIV_List_Active_Up);

	return 0;
}

DEFTMETHOD(PlaylistGroup_MoveDown)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	mygui->playlist->movedown_track(mygui->playlist, getv(data->LI_Playlist, MUIA_List_Active) + 1);

	DoMethod(data->LI_Playlist, MUIM_List_Move, MUIV_List_Move_Active, MUIV_List_Move_Next);

	set(data->LI_Playlist, MUIA_List_Active, MUIV_List_Active_Down);

	return 0;
}

DEFTMETHOD(PlaylistGroup_Shuffle)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	DoMethod(data->LI_Playlist, MUIM_List_Clear);

	mygui->playlist->shuffle_playlist(mygui->playlist);

	DoMethod(obj, MM_PlaylistGroup_Refresh, TRUE);

	return 0;
}

DEFSMETHOD(PlaylistGroup_Refresh)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);
	int i = 0;
/*
	if(guiIntfStruct.LengthInSec)
	{
		mygui->playlist->tracks[mygui->playlist->current]->duration = guiIntfStruct.LengthInSec;
	}
*/

	if(msg->rebuild)
	{
		set(data->LI_Playlist, MUIA_List_Quiet, TRUE);

		DoMethod(data->LI_Playlist, MUIM_List_Clear);

		for (i=0; i < mygui->playlist->trackcount; i++)
		{
			DoMethod(obj, MM_PlaylistGroup_Add, (APTR) mygui->playlist->tracks[i]);
		}
		set(data->LI_Playlist, MUIA_List_Active, mygui->playlist->current);

		set(data->LI_Playlist, MUIA_List_Quiet, FALSE);
	}
	else
	{
		/*
		if(mygui->playlist->tracks)
		{
			mygui->playlist->tracks[mygui->playlist->current]->duration = property_expand_string(guiIntfStruct.mpcontext, "${length}")
		}
		*/
		DoMethod(data->LI_Playlist, MUIM_List_Jump, mygui->playlist->current);
		DoMethod(data->LI_Playlist, MUIM_List_Redraw, MUIV_List_Redraw_All);
	}

	return 0;
}

DEFTMETHOD(PlaylistGroup_Clear)
{
	mygui->playlist->clear_playlist(mygui->playlist);

	DoMethod(obj, MUIM_List_Clear);

	return 0;
}

DEFTMETHOD(PlaylistGroup_Play)
{
	struct PlaylistGroupData *data = INST_DATA(cl, obj);

	if(getv(data->LI_Playlist, MUIA_List_Active) != MUIV_List_Active_Off)
	{
		set(obj, MA_PlaylistGroup_Current, getv(data->LI_Playlist, MUIA_List_Active));

		mplSetFileName(NULL, mygui->playlist->tracks[(mygui->playlist->current)]->filename, mygui->playlist->tracks[(mygui->playlist->current)]->type);
		mygui->playercontrol(evLoadPlay);
	}
	return 0;
}

DEFTMETHOD(PlaylistGroup_SaveRequester)
{
	APTR tags[] = { (APTR) ASLFR_TitleText, (APTR) "Save As...",
					(APTR) ASLFR_DoSaveMode, (APTR) TRUE,
					(APTR) ASLFR_InitialPattern, (APTR) PATTERN_PLAYLIST,
				    TAG_DONE };
	STRPTR p;
	TEXT path[1024];
	snprintf(path, sizeof(path), "%s", MorphOS_GetLastVisitedPath());
	AddPart(path, "playlist.pls", sizeof(path));

	set(app, MUIA_Application_Sleep, TRUE);

	p = asl_run(path, (APTR)&tags, FALSE);

	set(app, MUIA_Application_Sleep, FALSE);

	if (p)
	{
		int i;

		FILE * file = fopen(p, "w");

		if(file)
		{
			for (i=0; i < mygui->playlist->trackcount; i++)
			{
				fprintf(file, "%s\n",mygui->playlist->tracks[i]->filename);
			}
			fclose(file);
		}

		FreeVecTaskPooled(p);
	}

	return 0;
}

DEFSMETHOD(PlaylistGroup_Load)
{
	if(parse_filename(msg->path, playtree, mconfig, 0))
	{
		DoMethod(obj, MM_PlaylistGroup_Refresh, TRUE);
	}

	return 0;
}

DEFSMETHOD(PlaylistGroup_Save)
{
	if (msg->path)
	{
		int i;

		FILE * file = fopen(msg->path, "w");

		if(file)
		{
			for (i=0; i < mygui->playlist->trackcount; i++)
			{
				fprintf(file, "%s\n",mygui->playlist->tracks[i]->filename);
			}
			fclose(file);
		}
	}

	return 0;
}

BEGINMTABLE2(playlistgroupclass)
DECNEW(PlaylistGroup)
DECDISP(PlaylistGroup)
DECGET(PlaylistGroup)
DECSET(PlaylistGroup)
DECTMETHOD(PlaylistGroup_Play)
DECSMETHOD(PlaylistGroup_Add)
DECTMETHOD(PlaylistGroup_AddRequester)
DECTMETHOD(PlaylistGroup_Remove)
DECTMETHOD(PlaylistGroup_RemoveAll)
DECTMETHOD(PlaylistGroup_MoveUp)
DECTMETHOD(PlaylistGroup_MoveDown)
DECTMETHOD(PlaylistGroup_Shuffle)
DECTMETHOD(PlaylistGroup_Clear)
DECSMETHOD(PlaylistGroup_Refresh)
DECTMETHOD(PlaylistGroup_SaveRequester)
DECSMETHOD(PlaylistGroup_Load)
DECSMETHOD(PlaylistGroup_Save)
ENDMTABLE

DECSUBCLASS_NC(MUIC_Group, playlistgroupclass, PlaylistGroupData)
